#include "enigma.h"

namespace EnigmaLibrary
{
	/*
		getSeedOfRand μ
	*/
	void getSeedOfRand()
	{
		struct timeval tv;
		struct timezone tz;
		gettimeofday( &tv, &tz );

		::srand( tv.tv_sec );
	}

	/*
		vectorǤ򥹥֥뤹

		vector<unsigned int>& vec : ֥뤹vector
		unsigned int length : vectorνǿ
		unsigned int count : ֥
	*/
	void scramble( vector<unsigned int>& vec, unsigned int length, unsigned int count )
	{
		vector<unsigned int> numSeq;
		vec.clear();
		for ( unsigned int i = 0 ; i < length ; i++ ) {
			numSeq.push_back( i );
			vec.push_back( i );
		}

		for ( unsigned int i = 0 ; i < count ; i++ ) {
			if ( numSeq.size() < 2 ) break;
			unsigned int i1, i2;
			do {
				i1 = (unsigned int)( (double)rand() * numSeq.size() / ( (double)RAND_MAX + 1 ) );
				i2 = (unsigned int)( (double)rand() * numSeq.size() / ( (double)RAND_MAX + 1 ) );
			} while ( i1 == i2 );
			vec[numSeq[i1]] = numSeq[i2];
			vec[numSeq[i2]] = numSeq[i1];
			numSeq.erase( numSeq.begin() + ( ( i1 > i2 ) ? i1 : i2 ) );
			numSeq.erase( numSeq.begin() + ( ( i1 > i2 ) ? i2 : i1 ) );
		}
	}

	/*
		EnigmaIF 󥹥ȥ饯

		unsigned char start : ܡɤγʸ
		unsigned char end : ܡɤνλʸ

		ϰstartendǡԲǽʥɤϽ
	*/
	EnigmaIF::EnigmaIF( unsigned char start, unsigned char end )
	{
		for ( unsigned int c = start ; c <= end ; c++ )
			if ( isprint( c ) )
				key += (unsigned char)c;
	}

	/*
		EnigmaIF::keyIn ϥΥǥå֤

		unsigned char c : ϥ

		 : ϥΥǥå
						 (ĤʤäǸǤμΥǥå֤)
	*/
	unsigned int EnigmaIF::keyIn( unsigned char c ) const
	{
		unsigned int i = key.find( c );

		if ( i == string::npos )
			i = size();

		return( i );
	}

	/*
		EnigmaIF::print ¤ν
	*/
	void EnigmaIF::print() const
	{
		cout << "[Enigma I/F Keyboard]" << endl;
		cout << key << endl << endl;
	}

	/*
		Rotor 󥹥ȥ饯

		unsigned int length : ʸ
		unsigned int randomizeCount : 򥹥֥뤹
	*/
	Rotor::Rotor( unsigned int length, unsigned int randomizeCount )
		: location( 0 )
	{
		for ( unsigned int i = 0 ; i < length ; i++ )
			inShift.push_back( i );
		outShift.resize( length );

		for ( unsigned int i = 0 ; i < randomizeCount ; i++ ) {
			unsigned int i1 = (unsigned int)( (double)rand() * length / ( (double)RAND_MAX + 1 ) );
			unsigned int i2 = (unsigned int)( (double)rand() * length / ( (double)RAND_MAX + 1 ) );
			swap( inShift, i1, i2 );
		}

		for ( unsigned int i = 0 ; i < length ; i++ ) {
			int value = inShift[i];
			value -= i;
			if ( value < 0 ) value += length;
			inShift[i] = value;
			outShift[( i + value ) % length] = value;
		}
	}

	/*
		Rotor::encode Ź沽

		unsigned int i : ϰ

		 : ϰ(ϰϳξϰ֤򤽤Τޤ֤)
	*/
	unsigned int Rotor::encode( unsigned int i ) const
	{
		if ( i >= size() ) return( i );

		i += inShift[( i + location ) % size()];
		i %= size();

		return( i );
	}

	/*
		Rotor::decode 沽

		unsigned int i : ϰ

		 : ϰ(ϰϳξϰ֤򤽤Τޤ֤)
	*/
	unsigned int Rotor::decode( unsigned int i ) const
	{
		if ( i >= size() ) return( i );

		i += size() - outShift[( i + location ) % size()];
		i %= size();

		return( i );
	}

	/*
		Rotor::rotate ž

		 : ž˸ätrue֤
	*/
	bool Rotor::rotate()
	{
		location++;
		if ( location >= size() )
			location = 0;

		return( location == 0 );
	}

	/*
		Rotor::negRotate ղž

		 : 0ղžԤätrue֤
	*/
	bool Rotor::negRotate()
	{
		if ( location == 0 ) {
			location = size() - 1;
			return( true );
		} else {
			location--;
			return( false );
		}
	}

	/*
		Rotor::print ¤ν
	*/
	void Rotor::print() const
	{
		vector<unsigned int> o( size() );
		for ( unsigned int i = 0 ; i < size() ; i++ )
			o[encode( i )] = i;

		cout << "[Enigma Rotor]" << endl;
		cout << "<current location> = " << location << endl;

		cout << "<rotor>" << endl;
		for ( unsigned int i = 0 ; i < size() ; i++ )
			cout << "[" << setw( 2 ) << setfill( '0' ) << hex << o[i] << "]";
		cout << endl << endl;
	}

	/*
		Reflector 󥹥ȥ饯

		unsigned int length : ե쥯ʸ
	*/
	Reflector::Reflector( unsigned int length )
	{
		scramble( reflector, length, length / 2 );
	}

	/*
		Reflector::operator[] ϥǡФ֤ͤ

		unsigned int i : ϥǡ

		 : (ϰϳξϥǡ򤽤Τޤ֤)
	*/
	unsigned int Reflector::operator[]( unsigned int i ) const
	{
		if ( i >= size() ) return( i );

		return( reflector[i] );
	}

	/*
		Reflector::print ¤ν
	*/
	void Reflector::print() const
	{
		cout << "[Enigma Reflector]" << endl;
		for ( unsigned int i = 0 ; i < size() ; i++ )
			cout << "[" << setw( 3 ) << setfill( '0' ) << reflector[i] << "]";
		cout << endl;
	}

	/*
		PlugBoard::exchange ץ饰򴹤

		unsigned int i1, i2 : 򴹤
	*/
	bool PlugBoard::exchange( unsigned int i1, unsigned int i2 )
	{
		if ( i1 >= size() || i2 >= size() ) return( false );

		// ˸򴹤줿ץ饰ʤи᤹
		if ( plug[i1] != i1 ) {
			plug[plug[i1]] = plug[i1];
			plug[i1] = i1;
		}
		if ( plug[i2] != i2 ) {
			plug[plug[i2]] = plug[i2];
			plug[i2] = i2;
		}

		if ( i1 == i2 ) return( true );
		plug[i1] = i2;
		plug[i2] = i1;

		return( true );
	}

	/*
		PlugBoard::reset ץ饰ꥻåȤ
	*/
	void PlugBoard::reset()
	{
		for ( unsigned int i = 0 ; i < plug.size() ; i++ )
			exchange( i, i );
	}

	/*
		PlugBoard::print ¤ν
	*/
	void PlugBoard::print() const
	{
		cout << "[Enigma PlugBoard]" << endl;
		for ( unsigned int i = 0 ; i < plug.size() ; i++ ) {
			if ( i != plug[i] ) {
				cout << setw( 3 ) << setfill( '0' ) << i << "->";
				cout << setw( 3 ) << setfill( '0' ) << plug[i] << endl;
			}
		}
	}

	/*
		Enigma 󥹥ȥ饯

		unsigned char start : ܡɤγʸ
		unsigned char end : ܡɤνλʸ
		unsigned int rotorCount : ο
		unsigned int exchangeCount : ץ饰ܡɤθ򴹲
	*/
	Enigma::Enigma( unsigned char start, unsigned char end, unsigned int rotorCount, unsigned int exchangeCount )
	{
		getSeedOfRand();

		enigmaIF = new EnigmaIF( start, end );
		plugBoard = new PlugBoard( ioSize(), exchangeCount );
		for ( unsigned int i = 0 ; i < rotorCount ; i++ )
			vecRotor.push_back( new Rotor( ioSize() ) );
		reflector = new Reflector( ioSize() );
	}

	/*
		Enigma ԡ󥹥ȥ饯

		const Enigma& enigma : ԡ Enigma֥
	*/
	Enigma::Enigma( const Enigma& enigma )
	{
		enigmaIF = new EnigmaIF( *( enigma.enigmaIF ) );
		plugBoard = new PlugBoard( *( enigma.plugBoard ) );
		for ( unsigned int i = 0 ; i < enigma.rotorCount() ; i++ )
			vecRotor.push_back( new Rotor( *( ( enigma.vecRotor )[i] ) ) );
		reflector = new Reflector( *( enigma.reflector ) );
	}

	/*
		Enigma ǥȥ饯
	*/
	Enigma::~Enigma()
	{
		delete enigmaIF;
		delete plugBoard;
		for ( unsigned int i = 0 ; i < rotorCount() ; i++ )
			delete vecRotor[i];
		delete reflector;
	}

	/*
		Enigma::rotate βž

		ǸΥž0ä餽
		ž

		 : ǤƬˤ0äTrue֤
						 (Ĥޤ0ΰ֤ˤʤäTrue)
	*/
	bool Enigma::rotate()
	{
		for ( unsigned int i = rotorCount() ; i > 0 ; i-- )
			if ( ! ( vecRotor[i - 1] )->rotate() ) return( false );

		return( true );
	}

	/*
		Enigma::negRotate εղž

		ǸΥղž0η겼꤬ä餽
		ղž

		 : ǤƬˤǷ겼꤬äTrue֤
						 (Ĥޤ0ΰ֤ξ֤ǵղž򤷤True)
	*/
	bool Enigma::negRotate()
	{
		for ( unsigned int i = rotorCount() ; i > 0 ; i-- )
			if ( ! ( vecRotor[i - 1] )->negRotate() ) return( false );

		return( true );
	}

	/*
		Enigma::init ν
	*/
	void Enigma::init()
	{
		for ( unsigned int i = 0 ; i < rotorCount() ; i++ )
			( vecRotor[i] )->init();
	}

	/*
		Enigma::init ν

		const string& key : ˻Ѥ륭
	*/
	void Enigma::init( const string& key )
	{
		for ( unsigned int i = 0 ; i < rotorCount() ; i++ ) {
			unsigned int index = 0;
			if ( i < key.length() ) {
				unsigned int buff = keyIn( key[i] );
				if ( buff < ioSize() )
					index = buff;
			}
			( vecRotor[i] )->init( index );
		}
	}

	/*
		Enigma::current ƥΰ֤֤
	*/
	string Enigma::current() const
	{
		string s;
		for ( unsigned int i = 0 ; i < rotorCount() ; i++ ) {
			unsigned int index = ( vecRotor[i] )->current();
			unsigned char c = keyOut( index );
			s += c;
		}

		return( s );
	}

	/*
		Enigma::exchangeRotor θ

		unsigned int i1, i2 : 򴹤ֹ
	*/
	bool Enigma::exchangeRotor( unsigned int i1, unsigned int i2 )
	{
		if ( i1 >= rotorCount() || i2 >= rotorCount() ) return( false );

		Rotor* rotor = vecRotor[i1];
		vecRotor[i1] = vecRotor[i2];
		vecRotor[i2] = rotor;

		return( true );
	}

	/*
		Enigma::convert EnigmaʸŹ沽

		const string& inData : ʸ

		 : Ź沽ʸ
	*/
	string Enigma::convert( const string& inData )
	{
		string s;
		for ( unsigned int i = 0 ; i < inData.length() ; i++ ) {
			// ܡɤ
			unsigned int index = keyIn( inData[i] );
			if ( index >= ioSize() ) continue; // ϰϳʤ̵뤹

			// ץ饰ܡɤǤѴ
			index = plugBoard->encode( index );

			// ǤѴ
			for ( unsigned int i = 0 ; i < rotorCount() ; i++ ) {
				Rotor* p = vecRotor[i];
				index = p->encode( index );
			}

			// ե쥯ǤѴ
			index = ( *reflector )[index];

			// ǤεսѴ
			for ( unsigned int i = rotorCount() ; i > 0 ; i-- ) {
				Rotor* p = vecRotor[i - 1];
				index = p->decode( index );
			}

			// ץ饰ܡɤǤεѴ
			index = plugBoard->decode( index );

			// ʸν
			s += keyOut( index );

			rotate(); // ž
		}

		return( s );
	}

	/*
		Enigma::operator= 黻Ҥ¿

		const Enigma& enigma : ԡ Enigma֥

		 : ʬ
	*/
	Enigma& Enigma::operator=( const Enigma& enigma )
	{
		if ( this == &enigma ) return( *this ); //

		delete enigmaIF;
		delete plugBoard;
		for ( unsigned int i = 0 ; i < rotorCount() ; i++ )
			delete vecRotor[i];
		vecRotor.clear();
		delete reflector;

		enigmaIF = new EnigmaIF( *( enigma.enigmaIF ) );
		plugBoard = new PlugBoard( *( enigma.plugBoard ) );
		for ( unsigned int i = 0 ; i < enigma.rotorCount() ; i++ )
			vecRotor.push_back( new Rotor( *( ( enigma.vecRotor )[i] ) ) );
		reflector = new Reflector( *( enigma.reflector ) );

		return( *this );
	}

	/*
		Enigma::print Enigmaѥ᡼ɽ
	*/
	void Enigma::print() const
	{
		for ( unsigned int i = 0 ; i < rotorCount() ; i++ ) {
			Rotor* p = vecRotor[i];
			p->print();
		}
		plugBoard->print();
	}
}
